@@ -11,78 +11,102 @@ import asyncio
 import threading
 from abc import ABC, abstractmethod
 from types import TracebackType
-from typing import TYPE_CHECKING, AsyncIterator, Iterator, Generic, Union
-from typing import Any, Dict, List, Match, Optional, Sequence, Type, Tuple
+from typing import Any, AsyncIterator, Dict, Generic, Iterator, List, Match, IO
+from typing import Optional, Sequence, Tuple, Type, TypeVar, Union, TYPE_CHECKING
 
 from . import pq
+from . import adapt
 from . import errors as e
-from .pq import ExecStatus
 from .abc import Buffer, ConnectionType, PQGen, Transformer
-from .adapt import PyFormat
 from ._compat import create_task
 from ._cmodule import _psycopg
 from ._encodings import pgconn_encoding
 from .generators import copy_from, copy_to, copy_end
 
 if TYPE_CHECKING:
-    from .pq.abc import PGresult
     from .cursor import BaseCursor, Cursor
     from .cursor_async import AsyncCursor
     from .connection import Connection  # noqa: F401
     from .connection_async import AsyncConnection  # noqa: F401
 
+PY_TEXT = adapt.PyFormat.TEXT
+PY_BINARY = adapt.PyFormat.BINARY
+
 TEXT = pq.Format.TEXT
 BINARY = pq.Format.BINARY
 
+COPY_IN = pq.ExecStatus.COPY_IN
+COPY_OUT = pq.ExecStatus.COPY_OUT
+
+ACTIVE = pq.TransactionStatus.ACTIVE
+
+# Size of data to accumulate before sending it down the network. We fill a
+# buffer this size field by field, and when it passes the threshold size
+# we ship it, so it may end up being bigger than this.
+BUFFER_SIZE = 32 * 1024
+
+# Maximum data size we want to queue to send to the libpq copy. Sending a
+# buffer too big to be handled can cause an infinite loop in the libpq
+# (#255) so we want to split it in more digestable chunks.
+MAX_BUFFER_SIZE = 4 * BUFFER_SIZE
+# Note: making this buffer too large, e.g.
+# MAX_BUFFER_SIZE = 1024 * 1024
+# makes operations *way* slower! Probably triggering some quadraticity
+# in the libpq memory management and data sending.
+
+# Max size of the write queue of buffers. More than that copy will block
+# Each buffer should be around BUFFER_SIZE size.
+QUEUE_SIZE = 1024
+
 
 class BaseCopy(Generic[ConnectionType]):
     """
-    Base implementation for copy user interface
+    Base implementation for the copy user interface.
 
     Two subclasses expose real methods with the sync/async differences.
 
     The difference between the text and binary format is managed by two
     different `Formatter` subclasses.
 
-    While the interface doesn't dictate it, both subclasses are implemented
-    with a worker to perform I/O related work, consuming the data provided in
-    the correct format from a queue, while the main thread is concerned with
-    formatting the data in copy format and adding it to the queue.
+    Writing (the I/O part) is implemented in the subclasses by a `Writer` or
+    `AsyncWriter` instance. Normally writing implies sending copy data to a
+    database, but a different writer might be chosen, e.g. to stream data into
+    a file for later use.
     """
 
-    # Max size of the write queue of buffers. More than that copy will block
-    # Each buffer should be around BUFFER_SIZE size.
-    QUEUE_SIZE = 1024
-
-    # Size of data to accumulate before sending it down the network. We fill a
-    # buffer this size field by field, and when it passes the threshold size
-    # wee ship it, so it may end up being bigger than this.
-    BUFFER_SIZE = 32 * 1024
-
-    # Maximum data size we want to queue to send to the libpq copy. Sending a
-    # buffer too big to be handled can cause an infinite loop in the libpq
-    # (#255) so we want to split it in more digestable chunks.
-    MAX_BUFFER_SIZE = 4 * BUFFER_SIZE
-    # Note: making this buffer too large, e.g.
-    # MAX_BUFFER_SIZE = 1024 * 1024
-    # makes operations *way* slower! Probably triggering some quadraticity
-    # in the libpq memory management and data sending.
+    _Self = TypeVar("_Self", bound="BaseCopy[ConnectionType]")
 
     formatter: "Formatter"
 
-    def __init__(self, cursor: "BaseCursor[ConnectionType, Any]"):
+    def __init__(
+        self,
+        cursor: "BaseCursor[ConnectionType, Any]",
+        *,
+        binary: Optional[bool] = None,
+    ):
         self.cursor = cursor
         self.connection = cursor.connection
         self._pgconn = self.connection.pgconn
 
-        tx = cursor._tx
-        assert tx.pgresult, "The Transformer doesn't have a PGresult set"
-        self._pgresult: "PGresult" = tx.pgresult
-
-        if self._pgresult.binary_tuples == pq.Format.TEXT:
-            self.formatter = TextFormatter(tx, encoding=pgconn_encoding(self._pgconn))
+        result = cursor.pgresult
+        if result:
+            self._direction = result.status
+            if self._direction != COPY_IN and self._direction != COPY_OUT:
+                raise e.ProgrammingError(
+                    "the cursor should have performed a COPY operation;"
+                    f" its status is {pq.ExecStatus(self._direction).name} instead"
+                )
         else:
+            self._direction = COPY_IN
+
+        if binary is None:
+            binary = bool(result and result.binary_tuples)
+
+        tx: Transformer = getattr(cursor, "_tx", None) or adapt.Transformer(cursor)
+        if binary:
             self.formatter = BinaryFormatter(tx)
+        else:
+            self.formatter = TextFormatter(tx, encoding=pgconn_encoding(self._pgconn))
 
         self._finished = False
 
@@ -107,7 +131,7 @@ class BaseCopy(Generic[ConnectionType]):
 
         - On :sql:`COPY TO`, `!set_types()` allows to specify what types the
           operation returns. If `!set_types()` is not used, the data will be
-          reurned as unparsed strings or bytes instead of Python objects.
+          returned as unparsed strings or bytes instead of Python objects.
 
         - On :sql:`COPY FROM`, `!set_types()` allows to choose what type the
           database expects. This is especially useful in binary copy, because
@@ -117,7 +141,7 @@ class BaseCopy(Generic[ConnectionType]):
         registry = self.cursor.adapters.types
         oids = [t if isinstance(t, int) else registry.get_oid(t) for t in types]
 
-        if self._pgresult.status == ExecStatus.COPY_IN:
+        if self._direction == COPY_IN:
             self.formatter.transformer.set_dumper_types(oids, self.formatter.format)
         else:
             self.formatter.transformer.set_loader_types(oids, self.formatter.format)
@@ -134,6 +158,11 @@ class BaseCopy(Generic[ConnectionType]):
 
         # res is the final PGresult
         self._finished = True
+
+        # This result is a COMMAND_OK which has info about the number of rows
+        # returned, but not about the columns, which is instead an information
+        # that was received on the COPY_OUT result at the beginning of COPY.
+        # So, don't replace the results in the cursor, just update the rowcount.
         nrows = res.command_tuples
         self.cursor._rowcount = nrows if nrows is not None else -1
         return memoryview(b"")
@@ -152,25 +181,11 @@ class BaseCopy(Generic[ConnectionType]):
 
         return row
 
-    def _end_copy_in_gen(self, exc: Optional[BaseException]) -> PQGen[None]:
-        bmsg: Optional[bytes]
-        if exc:
-            msg = f"error from Python: {type(exc).__qualname__} - {exc}"
-            bmsg = msg.encode(pgconn_encoding(self._pgconn), "replace")
-        else:
-            bmsg = None
-
-        res = yield from copy_end(self._pgconn, bmsg)
-
-        nrows = res.command_tuples
-        self.cursor._rowcount = nrows if nrows is not None else -1
-        self._finished = True
-
     def _end_copy_out_gen(self, exc: Optional[BaseException]) -> PQGen[None]:
         if not exc:
             return
 
-        if self._pgconn.transaction_status != pq.TransactionStatus.ACTIVE:
+        if self._pgconn.transaction_status != ACTIVE:
             # The server has already finished to send copy data. The connection
             # is already in a good state.
             return
@@ -188,17 +203,40 @@ class BaseCopy(Generic[ConnectionType]):
 
 
 class Copy(BaseCopy["Connection[Any]"]):
-    """Manage a :sql:`COPY` operation."""
+    """Manage a :sql:`COPY` operation.
+
+    :param cursor: the cursor where the operation is performed.
+    :param binary: if `!True`, write binary format.
+    :param writer: the object to write to destination. If not specified, write
+        to the `!cursor` connection.
+
+    Choosing `!binary` is not necessary if the cursor has executed a
+    :sql:`COPY` operation, because the operation result describes the format
+    too. The parameter is useful when a `!Copy` object is created manually and
+    no operation is performed on the cursor, such as when using ``writer=``\\
+    `~psycopg.copy.FileWriter`.
+
+    """
 
     __module__ = "psycopg"
 
-    def __init__(self, cursor: "Cursor[Any]"):
-        super().__init__(cursor)
-        self._queue: queue.Queue[bytes] = queue.Queue(maxsize=self.QUEUE_SIZE)
-        self._worker: Optional[threading.Thread] = None
-        self._worker_error: Optional[BaseException] = None
+    writer: "Writer"
 
-    def __enter__(self) -> "Copy":
+    def __init__(
+        self,
+        cursor: "Cursor[Any]",
+        *,
+        binary: Optional[bool] = None,
+        writer: Optional["Writer"] = None,
+    ):
+        super().__init__(cursor, binary=binary)
+        if not writer:
+            writer = LibpqWriter(cursor)
+
+        self.writer = writer
+        self._write = writer.write
+
+    def __enter__(self: BaseCopy._Self) -> BaseCopy._Self:
         self._enter()
         return self
 
@@ -276,13 +314,87 @@ class Copy(BaseCopy["Connection[Any]"]):
         by exit. It is available if, despite what is documented, you end up
         using the `Copy` object outside a block.
         """
-        if self._pgresult.status == ExecStatus.COPY_IN:
-            self._write_end()
-            self.connection.wait(self._end_copy_in_gen(exc))
+        if self._direction == COPY_IN:
+            data = self.formatter.end()
+            if data:
+                self._write(data)
+            self.writer.finish(exc)
+            self._finished = True
         else:
             self.connection.wait(self._end_copy_out_gen(exc))
 
-    # Concurrent copy support
+
+class Writer(ABC):
+    """
+    A class to write copy data somewhere.
+    """
+
+    @abstractmethod
+    def write(self, data: Buffer) -> None:
+        """
+        Write some data to destination.
+        """
+        ...
+
+    def finish(self, exc: Optional[BaseException] = None) -> None:
+        """
+        Called when write operations are finished.
+
+        If operations finished with an error, it will be passed to ``exc``.
+        """
+        pass
+
+
+class LibpqWriter(Writer):
+    """
+    A `Writer` to write copy data to a Postgres database.
+    """
+
+    def __init__(self, cursor: "Cursor[Any]"):
+        self.cursor = cursor
+        self.connection = cursor.connection
+        self._pgconn = self.connection.pgconn
+
+    def write(self, data: Buffer) -> None:
+        if len(data) <= MAX_BUFFER_SIZE:
+            # Most used path: we don't need to split the buffer in smaller
+            # bits, so don't make a copy.
+            self.connection.wait(copy_to(self._pgconn, data))
+        else:
+            # Copy a buffer too large in chunks to avoid causing a memory
+            # error in the libpq, which may cause an infinite loop (#255).
+            for i in range(0, len(data), MAX_BUFFER_SIZE):
+                self.connection.wait(
+                    copy_to(self._pgconn, data[i : i + MAX_BUFFER_SIZE])
+                )
+
+    def finish(self, exc: Optional[BaseException] = None) -> None:
+        bmsg: Optional[bytes]
+        if exc:
+            msg = f"error from Python: {type(exc).__qualname__} - {exc}"
+            bmsg = msg.encode(pgconn_encoding(self._pgconn), "replace")
+        else:
+            bmsg = None
+
+        res = self.connection.wait(copy_end(self._pgconn, bmsg))
+        self.cursor._results = [res]
+
+
+class QueuedLibpqDriver(LibpqWriter):
+    """
+    A writer using a buffer to queue data to write to a Postgres database.
+
+    `write()` returns immediately, so that the main thread can be CPU-bound
+    formatting messages, while a worker thread can be IO-bound waiting to write
+    on the connection.
+    """
+
+    def __init__(self, cursor: "Cursor[Any]"):
+        super().__init__(cursor)
+
+        self._queue: queue.Queue[bytes] = queue.Queue(maxsize=QUEUE_SIZE)
+        self._worker: Optional[threading.Thread] = None
+        self._worker_error: Optional[BaseException] = None
 
     def worker(self) -> None:
         """Push data to the server when available from the copy queue.
@@ -302,7 +414,7 @@ class Copy(BaseCopy["Connection[Any]"]):
             # Propagate the error to the main thread.
             self._worker_error = ex
 
-    def _write(self, data: Buffer) -> None:
+    def write(self, data: Buffer) -> None:
         if not self._worker:
             # warning: reference loop, broken by _write_end
             self._worker = threading.Thread(target=self.worker)
@@ -313,20 +425,17 @@ class Copy(BaseCopy["Connection[Any]"]):
         if self._worker_error:
             raise self._worker_error
 
-        if len(data) <= self.MAX_BUFFER_SIZE:
+        if len(data) <= MAX_BUFFER_SIZE:
             # Most used path: we don't need to split the buffer in smaller
             # bits, so don't make a copy.
             self._queue.put(data)
         else:
             # Copy a buffer too large in chunks to avoid causing a memory
             # error in the libpq, which may cause an infinite loop (#255).
-            for i in range(0, len(data), self.MAX_BUFFER_SIZE):
-                self._queue.put(data[i : i + self.MAX_BUFFER_SIZE])
+            for i in range(0, len(data), MAX_BUFFER_SIZE):
+                self._queue.put(data[i : i + MAX_BUFFER_SIZE])
 
-    def _write_end(self) -> None:
-        data = self.formatter.end()
-        if data:
-            self._write(data)
+    def finish(self, exc: Optional[BaseException] = None) -> None:
         self._queue.put(b"")
 
         if self._worker:
@@ -337,18 +446,47 @@ class Copy(BaseCopy["Connection[Any]"]):
         if self._worker_error:
             raise self._worker_error
 
+        super().finish(exc)
+
+
+class FileWriter(Writer):
+    """
+    A `Writer` to write copy data to a file-like object.
+
+    :param file: the file where to write copy data. It must be open for writing
+        in binary mode.
+    """
+
+    def __init__(self, file: IO[bytes]):
+        self.file = file
+
+    def write(self, data: Buffer) -> None:
+        self.file.write(data)
+
 
 class AsyncCopy(BaseCopy["AsyncConnection[Any]"]):
     """Manage an asynchronous :sql:`COPY` operation."""
 
     __module__ = "psycopg"
 
-    def __init__(self, cursor: "AsyncCursor[Any]"):
-        super().__init__(cursor)
-        self._queue: asyncio.Queue[bytes] = asyncio.Queue(maxsize=self.QUEUE_SIZE)
-        self._worker: Optional[asyncio.Future[None]] = None
+    writer: "AsyncWriter"
 
-    async def __aenter__(self) -> "AsyncCopy":
+    def __init__(
+        self,
+        cursor: "AsyncCursor[Any]",
+        *,
+        binary: Optional[bool] = None,
+        writer: Optional["AsyncWriter"] = None,
+    ):
+        super().__init__(cursor, binary=binary)
+
+        if not writer:
+            writer = AsyncLibpqWriter(cursor)
+
+        self.writer = writer
+        self._write = writer.write
+
+    async def __aenter__(self: BaseCopy._Self) -> BaseCopy._Self:
         self._enter()
         return self
 
@@ -391,20 +529,85 @@ class AsyncCopy(BaseCopy["AsyncConnection[Any]"]):
             await self._write(data)
 
     async def finish(self, exc: Optional[BaseException]) -> None:
-        if self._pgresult.status == ExecStatus.COPY_IN:
-            await self._write_end()
-            await self.connection.wait(self._end_copy_in_gen(exc))
+        if self._direction == COPY_IN:
+            data = self.formatter.end()
+            if data:
+                await self._write(data)
+            await self.writer.finish(exc)
+            self._finished = True
         else:
             await self.connection.wait(self._end_copy_out_gen(exc))
 
-    # Concurrent copy support
+
+class AsyncWriter(ABC):
+    """
+    A class to write copy data somewhere (for async connections).
+    """
+
+    @abstractmethod
+    async def write(self, data: Buffer) -> None:
+        ...
+
+    async def finish(self, exc: Optional[BaseException] = None) -> None:
+        pass
+
+
+class AsyncLibpqWriter(AsyncWriter):
+    """
+    An `AsyncWriter` to write copy data to a Postgres database.
+    """
+
+    def __init__(self, cursor: "AsyncCursor[Any]"):
+        self.cursor = cursor
+        self.connection = cursor.connection
+        self._pgconn = self.connection.pgconn
+
+    async def write(self, data: Buffer) -> None:
+        if len(data) <= MAX_BUFFER_SIZE:
+            # Most used path: we don't need to split the buffer in smaller
+            # bits, so don't make a copy.
+            await self.connection.wait(copy_to(self._pgconn, data))
+        else:
+            # Copy a buffer too large in chunks to avoid causing a memory
+            # error in the libpq, which may cause an infinite loop (#255).
+            for i in range(0, len(data), MAX_BUFFER_SIZE):
+                await self.connection.wait(
+                    copy_to(self._pgconn, data[i : i + MAX_BUFFER_SIZE])
+                )
+
+    async def finish(self, exc: Optional[BaseException] = None) -> None:
+        bmsg: Optional[bytes]
+        if exc:
+            msg = f"error from Python: {type(exc).__qualname__} - {exc}"
+            bmsg = msg.encode(pgconn_encoding(self._pgconn), "replace")
+        else:
+            bmsg = None
+
+        res = await self.connection.wait(copy_end(self._pgconn, bmsg))
+        self.cursor._results = [res]
+
+
+class AsyncQueuedLibpqWriter(AsyncLibpqWriter):
+    """
+    An `AsyncWriter` using a buffer to queue data to write.
+
+    `write()` returns immediately, so that the main thread can be CPU-bound
+    formatting messages, while a worker thread can be IO-bound waiting to write
+    on the connection.
+    """
+
+    def __init__(self, cursor: "AsyncCursor[Any]"):
+        super().__init__(cursor)
+
+        self._queue: asyncio.Queue[bytes] = asyncio.Queue(maxsize=QUEUE_SIZE)
+        self._worker: Optional[asyncio.Future[None]] = None
 
     async def worker(self) -> None:
         """Push data to the server when available from the copy queue.
 
         Terminate reading when the queue receives a false-y value.
 
-        The function is designed to be run in a separate thread.
+        The function is designed to be run in a separate task.
         """
         while True:
             data = await self._queue.get()
@@ -412,30 +615,29 @@ class AsyncCopy(BaseCopy["AsyncConnection[Any]"]):
                 break
             await self.connection.wait(copy_to(self._pgconn, data))
 
-    async def _write(self, data: Buffer) -> None:
+    async def write(self, data: Buffer) -> None:
         if not self._worker:
             self._worker = create_task(self.worker())
 
-        if len(data) <= self.MAX_BUFFER_SIZE:
+        if len(data) <= MAX_BUFFER_SIZE:
             # Most used path: we don't need to split the buffer in smaller
             # bits, so don't make a copy.
             await self._queue.put(data)
         else:
             # Copy a buffer too large in chunks to avoid causing a memory
             # error in the libpq, which may cause an infinite loop (#255).
-            for i in range(0, len(data), self.MAX_BUFFER_SIZE):
-                await self._queue.put(data[i : i + self.MAX_BUFFER_SIZE])
+            for i in range(0, len(data), MAX_BUFFER_SIZE):
+                await self._queue.put(data[i : i + MAX_BUFFER_SIZE])
 
-    async def _write_end(self) -> None:
-        data = self.formatter.end()
-        if data:
-            await self._write(data)
+    async def finish(self, exc: Optional[BaseException] = None) -> None:
         await self._queue.put(b"")
 
         if self._worker:
             await asyncio.gather(self._worker)
             self._worker = None  # break reference loops if any
 
+        await super().finish(exc)
+
 
 class Formatter(ABC):
     """
@@ -443,7 +645,6 @@ class Formatter(ABC):
     """
 
     format: pq.Format
-    BUFFER_SIZE = BaseCopy.BUFFER_SIZE
 
     def __init__(self, transformer: Transformer):
         self.transformer = transformer
@@ -469,7 +670,7 @@ class Formatter(ABC):
 
 class TextFormatter(Formatter):
 
-    format = pq.Format.TEXT
+    format = TEXT
 
     def __init__(self, transformer: Transformer, encoding: str = "utf-8"):
         super().__init__(transformer)
@@ -492,7 +693,7 @@ class TextFormatter(Formatter):
         self._row_mode = True
 
         format_row_text(row, self.transformer, self._write_buffer)
-        if len(self._write_buffer) > self.BUFFER_SIZE:
+        if len(self._write_buffer) > BUFFER_SIZE:
             buffer, self._write_buffer = self._write_buffer, bytearray()
             return buffer
         else:
@@ -514,7 +715,7 @@ class TextFormatter(Formatter):
 
 class BinaryFormatter(Formatter):
 
-    format = pq.Format.BINARY
+    format = BINARY
 
     def __init__(self, transformer: Transformer):
         super().__init__(transformer)
@@ -549,7 +750,7 @@ class BinaryFormatter(Formatter):
             self._signature_sent = True
 
         format_row_binary(row, self.transformer, self._write_buffer)
-        if len(self._write_buffer) > self.BUFFER_SIZE:
+        if len(self._write_buffer) > BUFFER_SIZE:
             buffer, self._write_buffer = self._write_buffer, bytearray()
             return buffer
         else:
@@ -597,7 +798,7 @@ def _format_row_text(
 
     for item in row:
         if item is not None:
-            dumper = tx.get_dumper(item, PyFormat.TEXT)
+            dumper = tx.get_dumper(item, PY_TEXT)
             b = dumper.dump(item)
             out += _dump_re.sub(_dump_sub, b)
         else:
@@ -616,7 +817,7 @@ def _format_row_binary(
         out = bytearray()
 
     out += _pack_int2(len(row))
-    adapted = tx.dump_sequence(row, [PyFormat.BINARY] * len(row))
+    adapted = tx.dump_sequence(row, [PY_BINARY] * len(row))
     for b in adapted:
         if b is not None:
             out += _pack_int4(len(b))
